"""Utilities for smoothing geometries.

Fork + Port of <https://github.com/philipschall/shapelysmooth> (Public domain)
"""

from __future__ import annotations

from typing import TYPE_CHECKING, TypeAlias

import numpy as np
from shapely.geometry import LineString, Polygon
from typing_extensions import assert_never

if TYPE_CHECKING:
    from numpy.typing import NDArray

T: TypeAlias = LineString | Polygon


class InputeTypeError(TypeError):
    """Raised when the input geometry is of the incorrect type."""


def catmull_rom_smooth(geometry: T, alpha: float = 0.5, subdivs: int = 10) -> T:
    """Polyline smoothing using Catmull-Rom splines.

    Args:
        geometry: The geometry to smooth
        alpha: The tension parameter, between 0 and 1 inclusive. Defaults to 0.5.
               - For uniform Catmull-Rom splines, alpha = 0.
               - For centripetal Catmull-Rom splines, alpha = 0.5.
               - For chordal Catmull-Rom splines, alpha = 1.0.
        subdivs:
            Number of subdivisions of each polyline segment. Default value: 10.

    Returns: The smoothed geometry.
    """
    coords, interior_coords = _get_coords(geometry)
    coords_smoothed = _catmull_rom(coords, alpha=alpha, subdivs=subdivs)
    if isinstance(geometry, LineString):
        return type(geometry)(coords_smoothed)
    elif isinstance(geometry, Polygon):
        interior_coords_smoothed = [
            _catmull_rom(c, alpha=alpha, subdivs=subdivs) for c in interior_coords
        ]
        return type(geometry)(coords_smoothed, holes=interior_coords_smoothed)
    else:
        assert_never(geometry)


def _catmull_rom(
    coords: NDArray,
    *,
    alpha: float = 0.5,
    subdivs: int = 8,
) -> list[tuple[float, float]]:
    arr = np.asarray(coords, dtype=float)
    if arr.shape[0] < 2:
        return arr.tolist()

    is_closed = np.allclose(arr[0], arr[-1])
    if is_closed:
        arr = np.vstack([arr[-2], arr, arr[2]])
    else:
        arr = np.vstack(
            [
                2.0 * arr[0] + 1.0 * arr[1],
                arr,
                2.0 * arr[-1] + 0.0 * arr[-2],
            ]
        )

    new_ls = [tuple(arr[1])]
    for k in range(len(arr) - 3):
        slice4 = arr[k : k + 4]
        tangents = [0.0]
        for j in range(3):
            dist = float(np.linalg.norm(slice4[j + 1] - slice4[j]))
            tangents.append(float(tangents[-1] + dist**alpha))

        # Resample: subdivs-1 samples strictly between t1 and t2
        seg_len = (tangents[2] - tangents[1]) / float(subdivs)
        if subdivs > 1:
            ts = np.linspace(tangents[1] + seg_len, tangents[2] - seg_len, subdivs - 1)
        else:
            ts = np.array([])

        interpolants = _recursive_eval(slice4, tangents, ts)
        new_ls.extend(interpolants)
        new_ls.append(tuple(slice4[2]))
    return new_ls


def _recursive_eval(
    slice4: NDArray, tangents: list[float], ts: NDArray
) -> list[tuple[float, float]]:
    """De Boor/De Casteljau-style recursive linear interpolation over 4 control points.

    Parameterized by the non-uniform 'tangents' values.
    """
    # N.B. comments are LLM-generated

    out = []
    for tp in ts:
        # Start with the 4 control points for this segment
        points = slice4.copy()
        # Perform 3 levels of linear interpolation (De Casteljau's algorithm)
        for r in range(1, 4):
            idx = max(r - 2, 0)
            new_points = []
            # Interpolate between points at this level
            for i in range(4 - r):
                # Compute denominator for parameterization
                denom = tangents[i + r - idx] - tangents[i + idx]
                if denom == 0:
                    # If degenerate (coincident tangents), use midpoint
                    left_w = right_w = 0.5
                else:
                    # Otherwise, compute weights for linear interpolation
                    left_w = (tangents[i + r - idx] - tp) / denom
                    right_w = (tp - tangents[i + idx]) / denom
                # Weighted average of the two points
                pt = left_w * points[i] + right_w * points[i + 1]
                new_points.append(pt)
            # Move to the next level with the new set of points
            points = np.array(new_points)
        # The final point is the interpolated value for this parameter tp
        out.append(tuple(points[0]))
    return out


def _get_coords(
    geometry: LineString | Polygon,
) -> tuple[NDArray, list[NDArray]]:
    if isinstance(geometry, LineString):
        return np.array(geometry.coords), []
    elif isinstance(geometry, Polygon):
        return np.array(geometry.exterior.coords), [
            np.array(hole.coords) for hole in geometry.interiors
        ]
    else:
        assert_never(geometry)



import numpy as np
import pytest
from affine import Affine
from pydantic import ValidationError
from pyproj.crs.crs import CRS

from rastr.raster import Raster, RasterMeta

_NZTM_CRS = CRS.from_epsg(2193)


class TestRaster:
    def test_instantiable(self):
        """Test that Raster can be instantiated."""
        arr = np.array([[1, 2], [3, 4]])
        meta = RasterMeta(cell_size=1, crs=_NZTM_CRS, transform=Affine.scale(1.0, 1.0))
        raster = Raster(arr=arr, raster_meta=meta)
        assert isinstance(raster, Raster)

    def test_3d_fails(self):
        """Test that Raster cannot be instantiated with a 3D array."""
        arr = np.array([[[1, 1], [2, 2]], [[3, 3], [4, 4]]])  # 2 x 2 x 2 array
        meta = RasterMeta(cell_size=1, crs=_NZTM_CRS, transform=Affine.scale(1.0, 1.0))

        with pytest.raises(ValidationError):
            Raster(arr=arr, raster_meta=meta)

    def test_2x2x1_fails(self):
        arr = np.array([[[0.23465047], [0.77642868]], [[0.92393235], [0.55804058]]])
        meta = RasterMeta(cell_size=1, crs=_NZTM_CRS, transform=Affine.scale(1.0, 1.0))
        with pytest.raises(ValidationError):
            Raster(arr=arr, raster_meta=meta)

    class TestExample:
        def test_example(self):
            # Arrange
            arr = np.array([[1, 2], [3, 4]])
            meta = RasterMeta(
                cell_size=1,
                crs=_NZTM_CRS,
                transform=Affine.scale(1.0, 1.0),
            )

            # Act
            example_meta = Raster(
                arr=arr,
                raster_meta=meta,
            )

            # Assert
            assert isinstance(example_meta, Raster)

    def test_get_cell_centre_coords(self):
        # Arrange
        cell_size = 2.0
        crs = _NZTM_CRS
        transform = Affine.translation(100, 200) * Affine.scale(cell_size, -cell_size)
        meta = RasterMeta(cell_size=cell_size, crs=crs, transform=transform)
        shape = (2, 3)

        # Act
        coords = meta.get_cell_centre_coords(shape)

        # Assert
        assert coords.shape == (2, 3, 2)
        expected = np.array(
            [
                [[101.0, 199.0], [103.0, 199.0], [105.0, 199.0]],
                [[101.0, 197.0], [103.0, 197.0], [105.0, 197.0]],
            ]
        )
        np.testing.assert_allclose(coords, expected)



from __future__ import annotations

import importlib.util
import warnings
from typing import TYPE_CHECKING, TypeVar

import numpy as np
import rasterio.features
import rasterio.transform
from affine import Affine
from pyproj import CRS
from shapely.geometry import Point

from rastr.gis.fishnet import create_point_grid, get_point_grid_shape
from rastr.meta import RasterMeta
from rastr.raster import Raster

if TYPE_CHECKING:
    from collections.abc import Iterable

    import geopandas as gpd
    from numpy.typing import ArrayLike
    from shapely.geometry import Polygon


TQDM_INSTALLED = importlib.util.find_spec("tqdm") is not None

_T = TypeVar("_T")


class MissingColumnsError(ValueError):
    """Raised when target columns are missing from the GeoDataFrame."""


class NonNumericColumnsError(ValueError):
    """Raised when target columns contain non-numeric data."""


class RasterizationError(ValueError):
    """Base exception for rasterization errors."""


class OverlappingGeometriesError(RasterizationError):
    """Raised when geometries overlap, which could lead to data loss."""


def raster_distance_from_polygon(
    polygon: Polygon,
    *,
    raster_meta: RasterMeta,
    extent_polygon: Polygon | None = None,
    snap_raster: Raster | None = None,
    show_pbar: bool = False,
) -> Raster:
    """Make a raster where each cell's value is its centre's distance to a polygon.

    The raster should use a projected coordinate system.

    Parameters:
        polygon: Polygon to measure distances to.
        raster_meta: Raster configuration (giving cell_size, CRS, etc.).
        extent_polygon: Polygon for raster cell extent; The bounding box of this
                        polygon is the bounding box of the output raster. Cells outside
                        this polygon but within the bounding box will be NaN-valued, and
                        cells will not be generated centred outside the bounding box of
                        this polygon.
        snap_raster: An alternative to using the extent_polygon. If provided, the raster
                     must have the exact same cell alignment as the snap_raster.
        show_pbar: Whether to show a progress bar during the distance calculation.

    Returns:
        Array storing the distance between cell centres and the polygon. Cell are
        NaN-valued if they are within the polygon or outside the extent polygon.

    Raises:
        ValueError: If the provided CRS is geographic (lat/lon).
    """
    if show_pbar and not TQDM_INSTALLED:
        msg = "The 'tqdm' package is not installed. Progress bars will not be shown."
        warnings.warn(msg, UserWarning, stacklevel=2)
        show_pbar = False

    # Check if the provided CRS is projected (cartesian)
    if raster_meta.crs.is_geographic:
        err_msg = (
            "The provided CRS is geographic (lat/lon). Please use a projected CRS."
        )
        raise ValueError(err_msg)

    if extent_polygon is None and snap_raster is None:
        err_msg = "Either 'extent_polygon' or 'snap_raster' must be provided. "
        raise ValueError(err_msg)
    elif extent_polygon is not None and snap_raster is not None:
        err_msg = "Only one of 'extent_polygon' or 'snap_raster' can be provided. "
        raise ValueError(err_msg)
    elif extent_polygon is None and snap_raster is not None:
        # Calculate the coordinates
        x, y = snap_raster.get_xy()

        # Create a mask to identify points for which distance should be calculated
        distance_extent = snap_raster.bbox.difference(polygon)
    elif extent_polygon is not None and snap_raster is None:
        x, y = create_point_grid(
            bounds=extent_polygon.bounds, cell_size=raster_meta.cell_size
        )
        distance_extent = extent_polygon.difference(polygon)
    else:
        raise AssertionError

    pts = [Point(x, y) for x, y in zip(x.flatten(), y.flatten(), strict=True)]

    _pts = _pbar(pts, desc="Finding points within extent") if show_pbar else pts
    mask = [distance_extent.intersects(pt) for pt in _pts]

    _pts = _pbar(pts, desc="Calculating distances") if show_pbar else pts
    distances = np.where(mask, np.array([polygon.distance(pt) for pt in _pts]), np.nan)
    distance_raster = distances.reshape(x.shape)

    return Raster(arr=distance_raster, raster_meta=raster_meta)


def _pbar(iterable: Iterable[_T], *, desc: str | None = None) -> Iterable[_T]:
    from tqdm import tqdm

    return tqdm(iterable, desc=desc)


def full_raster(
    raster_meta: RasterMeta,
    *,
    bounds: tuple[float, float, float, float],
    fill_value: float = np.nan,
) -> Raster:
    """Create a raster with a specified fill value for all cells."""
    shape = get_point_grid_shape(bounds=bounds, cell_size=raster_meta.cell_size)
    arr = np.full(shape, fill_value, dtype=np.float32)
    return Raster(arr=arr, raster_meta=raster_meta)


def rasterize_gdf(
    gdf: gpd.GeoDataFrame,
    *,
    raster_meta: RasterMeta,
    target_cols: list[str],
) -> list[Raster]:
    """Rasterize geometries from a GeoDataFrame.

    Supports polygons, points, linestrings, and other geometry types.
    Gaps will be set as NaN.

    Args:
        gdf: The geometries to rasterize (polygons, points, linestrings, etc.).
        raster_meta: Metadata for the created rasters.
        target_cols: A list of columns from the GeoDataFrame containing numeric
                     datatypes. Each column will correspond to a separate raster
                     in the output.

    Returns:
        Rasters for each column in `target_cols`.

    Raises:
        MissingColumnsError: If any of the target columns are not found in the
                             GeoDataFrame.
        NonNumericColumnsError: If any of the target columns contain non-numeric data.
        OverlappingGeometriesError: If any geometries overlap, which could lead to
                                    data loss in the rasterization process.
    """
    # Validate inputs using helper functions
    _validate_columns_exist(gdf, target_cols)
    _validate_columns_numeric(gdf, target_cols)
    _validate_no_overlapping_geometries(gdf)

    # Get the bounds from the GeoDataFrame and expand them to include potential gaps
    bounds = gdf.total_bounds
    min_x, min_y, max_x, max_y = bounds
    cell_size = raster_meta.cell_size

    # Expand bounds by at least one cell size to ensure there are potential gaps
    buffer = cell_size
    expanded_bounds = (min_x - buffer, min_y - buffer, max_x + buffer, max_y + buffer)

    # Create point grid to get raster dimensions and transform
    shape = get_point_grid_shape(bounds=expanded_bounds, cell_size=cell_size)

    # Create the affine transform for rasterization
    transform = Affine.translation(
        expanded_bounds[0], expanded_bounds[3]
    ) * Affine.scale(cell_size, -cell_size)

    # Create rasters for each target column using rasterio.features.rasterize
    rasters = []
    for col in target_cols:
        # Create (geometry, value) pairs for rasterization
        shapes = [
            (geom, value) for geom, value in zip(gdf.geometry, gdf[col], strict=True)
        ]

        # Rasterize the geometries with their values
        raster_array = rasterio.features.rasterize(
            shapes,
            out_shape=shape,
            transform=transform,
            # Fill gaps with NaN
            fill=np.nan,  # type: ignore[reportArgumentType] docstring contradicts inferred annotation
            dtype=np.float32,
        )

        # Create Raster
        raster = Raster(arr=raster_array, raster_meta=raster_meta)
        rasters.append(raster)

    return rasters


def _validate_columns_exist(gdf: gpd.GeoDataFrame, target_cols: list[str]) -> None:
    """Validate that all target columns exist in the GeoDataFrame.

    Args:
        gdf: The GeoDataFrame to check.
        target_cols: List of column names to validate.

    Raises:
        MissingColumnsError: If any columns are missing.
    """
    missing_cols = [col for col in target_cols if col not in gdf.columns]
    if missing_cols:
        msg = f"Target columns not found in GeoDataFrame: {missing_cols}"
        raise MissingColumnsError(msg)


def _validate_columns_numeric(gdf: gpd.GeoDataFrame, target_cols: list[str]) -> None:
    """Validate that all target columns contain numeric data.

    Args:
        gdf: The GeoDataFrame to check.
        target_cols: List of column names to validate.

    Raises:
        NonNumericColumnsError: If any columns contain non-numeric data.
    """
    import pandas as pd

    non_numeric_cols = []
    for col in target_cols:
        if not pd.api.types.is_numeric_dtype(gdf[col]):
            non_numeric_cols.append(col)
    if non_numeric_cols:
        msg = f"Target columns must contain numeric data: {non_numeric_cols}"
        raise NonNumericColumnsError(msg)


def _validate_no_overlapping_geometries(gdf: gpd.GeoDataFrame) -> None:
    """Validate that geometries do not overlap.

    Args:
        gdf: The GeoDataFrame to check for overlapping geometries.

    Raises:
        OverlappingGeometriesError: If any geometries overlap.
    """
    # Check for overlaps by testing each geometry against all others
    geometries = gdf.geometry.to_numpy()

    for i in range(len(geometries)):
        for j in range(i + 1, len(geometries)):
            geom_i = geometries[i]
            geom_j = geometries[j]

            # Skip invalid geometries
            if not geom_i.is_valid or not geom_j.is_valid:
                continue

            # Check if geometries overlap (not just touch)
            if geom_i.overlaps(geom_j):
                msg = (
                    f"Overlapping geometries detected at indices {i} and {j}. "
                    "Overlapping geometries can lead to data loss during rasterization."
                )
                raise OverlappingGeometriesError(msg)


def raster_from_point_cloud(
    x: ArrayLike,
    y: ArrayLike,
    z: ArrayLike,
    *,
    crs: CRS | str,
    cell_size: float | None = None,
) -> Raster:
    """Create a raster from a point cloud via interpolation.

    Interpolation is only possible within the convex hull of the points. Outside of
    this, cells will be NaN-valued.

    All (x,y) points must be unique.

    Args:
        x: X coordinates of points.
        y: Y coordinates of points.
        z: Values at each (x, y) point to assign the raster.
        crs: Coordinate reference system for the (x, y) coordinates.
        cell_size: Desired cell size for the raster. If None, a heuristic is used based
                   on the spacing between (x, y) points.

    Returns:
        Raster containing the interpolated values.

    Raises:
        ValueError: If any (x, y) points are duplicated, or if they are all collinear.
    """
    from scipy.interpolate import LinearNDInterpolator
    from scipy.spatial import KDTree, QhullError

    x = np.asarray(x).ravel()
    y = np.asarray(y).ravel()
    z = np.asarray(z).ravel()
    crs = CRS.from_user_input(crs)

    # Validate input arrays
    if len(x) != len(y) or len(x) != len(z):
        msg = "Length of x, y, and z must be equal."
        raise ValueError(msg)
    xy_finite_mask = np.isfinite(x) & np.isfinite(y)
    if np.any(~xy_finite_mask):
        msg = "Some (x,y) points are NaN-valued or non-finite. These will be ignored."
        warnings.warn(msg, stacklevel=2)
        x = x[xy_finite_mask]
        y = y[xy_finite_mask]
        z = z[xy_finite_mask]
    if len(x) < 3:
        msg = (
            "At least three valid (x, y, z) points are required to triangulate a "
            "surface."
        )
        raise ValueError(msg)
    # Check for duplicate (x, y) points
    xy_points = np.column_stack((x, y))
    if len(xy_points) != len(np.unique(xy_points, axis=0)):
        msg = "Duplicate (x, y) points found. Each (x, y) point must be unique."
        raise ValueError(msg)

    # Heuristic for cell size if not provided
    if cell_size is None:
        # Half the 5th percentile of nearest neighbor distances between the (x,y) points
        tree = KDTree(xy_points)
        distances, _ = tree.query(xy_points, k=2)
        distances: np.ndarray
        cell_size = float(np.percentile(distances[distances > 0], 5)) / 2

    # Compute bounds from data
    minx, miny, maxx, maxy = np.min(x), np.min(y), np.max(x), np.max(y)

    # Compute grid shape
    width = int(np.ceil((maxx - minx) / cell_size))
    height = int(np.ceil((maxy - miny) / cell_size))
    shape = (height, width)

    # Compute transform: upper left corner is (minx, maxy)
    transform = Affine.translation(minx, maxy) * Affine.scale(cell_size, -cell_size)

    # Create grid coordinates for raster cells
    rows, cols = np.indices(shape)
    xs, ys = rasterio.transform.xy(
        transform=transform, rows=rows, cols=cols, offset="center"
    )
    grid_x = np.array(xs).ravel()
    grid_y = np.array(ys).ravel()

    # Perform interpolation
    try:
        interpolator = LinearNDInterpolator(
            points=xy_points, values=z, fill_value=np.nan
        )
    except QhullError as err:
        msg = (
            "Failed to interpolate. This may be due to insufficient or "
            "degenerate input points. Ensure that the (x, y) points are not all "
            "collinear (i.e. that the convex hull is non-degenerate)."
        )
        raise ValueError(msg) from err

    grid_values = np.array(interpolator(np.column_stack((grid_x, grid_y))))

    arr = grid_values.reshape(shape).astype(np.float32)

    raster_meta = RasterMeta(
        cell_size=cell_size,
        crs=crs,
        transform=transform,
    )
    return Raster(arr=arr, raster_meta=raster_meta)



from pathlib import Path

import pytest

collect_ignore_glob = ["assets/**"]
pytest_plugins = []


@pytest.fixture(scope="session")
def assets_dir() -> Path:
    """Return a path to the test assets directory."""
    return Path(__file__).parent / "assets"



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from shapely.geometry import Polygon

from rastr.gis.fishnet import (
    create_fishnet,
    create_point_grid,
    get_point_grid_shape,
)

if TYPE_CHECKING:
    from collections.abc import Collection

    from geopandas.array import GeometryArray


def _normalize_polygon(polygon: Polygon) -> Polygon:
    """Normalize a polygon's coordinates by sorting them."""
    coords = sorted(
        polygon.exterior.coords[:-1]
    )  # Sort the coordinates, exclude the closing point
    return Polygon(coords)


def assert_polygons_equal(
    result: Collection[Polygon] | GeometryArray,
    expected: Collection[Polygon] | GeometryArray,
    tol: float = 1e-9,
) -> None:
    """Assert that two lists of polygons are equal."""
    result = list(result)
    expected = list(expected)

    if len(result) != len(expected):
        pytest.fail(f"Expected {len(expected)} geometries, got {len(result)}")

    normalized_result = [_normalize_polygon(p) for p in result]
    normalized_expected = [_normalize_polygon(p) for p in expected]

    for res_poly in normalized_result:
        if not any(
            res_poly.equals_exact(exp_poly, tol) for exp_poly in normalized_expected
        ):
            pytest.fail(f"Geometry {res_poly} not found in expected geometries")


class TestFishnet:
    def test_regular_example(self):
        bounds = (0.0, 0.0, 2.0, 2.0)
        expected_polygons = [
            Polygon([(0, 0), (1, 0), (1, 1), (0, 1)]),
            Polygon([(1, 0), (2, 0), (2, 1), (1, 1)]),
            Polygon([(0, 1), (1, 1), (1, 2), (0, 2)]),
            Polygon([(1, 1), (2, 1), (2, 2), (1, 2)]),
        ]

        result = create_fishnet(bounds=bounds, res=1.0)

        assert_polygons_equal(result, expected_polygons)

    def test_res_not_factor_of_bounds_dims(self):
        bounds = (0.0, 0.0, 1.0, 2.0)
        expected_polygons = [
            Polygon([(0, 2), (0.6, 2), (0.6, 1.4), (0, 1.4)]),
            Polygon([(0.6, 2), (1.2, 2), (1.2, 1.4), (0.6, 1.4)]),
            Polygon([(0, 1.4), (0.6, 1.4), (0.6, 0.8), (0, 0.8)]),
            Polygon([(0.6, 1.4), (1.2, 1.4), (1.2, 0.8), (0.6, 0.8)]),
            Polygon([(0, 0.8), (0.6, 0.8), (0.6, 0.2), (0, 0.2)]),
            Polygon([(0.6, 0.8), (1.2, 0.8), (1.2, 0.2), (0.6, 0.2)]),
            Polygon([(0, 0.2), (0.6, 0.2), (0.6, -0.4), (0, -0.4)]),
            Polygon([(0.6, 0.2), (1.2, 0.2), (1.2, -0.4), (0.6, -0.4)]),
        ]
        result = create_fishnet(bounds=bounds, res=0.6)

        assert_polygons_equal(result, expected_polygons)


class TestGetPointGridShape:
    def test_matches_create_point_grid_simple(self):
        bounds = (0.0, 0.0, 2.0, 2.0)
        cell_size = 1.0
        expected_shape = create_point_grid(bounds=bounds, cell_size=cell_size)[0].shape

        shape = get_point_grid_shape(bounds=bounds, cell_size=cell_size)

        assert shape == expected_shape

    def test_matches_create_point_grid_borderline(self):
        bounds = (0.0, 0.0, 2.0, 2.0)
        cell_size = 1.5
        expected_shape = create_point_grid(bounds=bounds, cell_size=cell_size)[0].shape

        shape = get_point_grid_shape(bounds=bounds, cell_size=cell_size)

        assert shape == expected_shape



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
import rasterio
import rasterio.merge
from pyproj.crs.crs import CRS

from rastr.meta import RasterMeta
from rastr.raster import Raster

if TYPE_CHECKING:
    from numpy.typing import NDArray


def read_raster_inmem(
    raster_path: Path | str, *, crs: CRS | str | None = None
) -> Raster:
    """Read raster data from a file and return an in-memory Raster object."""
    crs = CRS.from_user_input(crs) if crs is not None else None

    with rasterio.open(raster_path, mode="r") as dst:
        # Read the entire array
        arr: NDArray[np.float64] = dst.read()
        arr = arr.squeeze().astype(np.float64)
        # Extract metadata
        cell_size = dst.res[0]
        if crs is None:
            crs = CRS.from_user_input(dst.crs)
        transform = dst.transform
        nodata = dst.nodata
        if nodata is not None:
            arr[arr == nodata] = np.nan

    raster_meta = RasterMeta(cell_size=cell_size, crs=crs, transform=transform)
    raster_obj = Raster(arr=arr, raster_meta=raster_meta)
    return raster_obj


def read_raster_mosaic_inmem(
    mosaic_dir: Path | str, *, glob: str = "*.tif", crs: CRS | None = None
) -> Raster:
    """Read a raster mosaic from a directory and return an in-memory Raster object.

    This assumes that all rasters have the same metadata, e.g. coordinate system,
    cell size, etc.
    """
    mosaic_dir = Path(mosaic_dir)
    raster_paths = list(mosaic_dir.glob(glob))
    if not raster_paths:
        msg = f"No raster files found in {mosaic_dir} matching {glob}"
        raise FileNotFoundError(msg)

    # Sort raster_paths in alphabetical order by stem
    raster_paths.sort(key=lambda p: p.stem)

    # Open all TIFF datasets using context managers to ensure proper closure
    sources = []
    try:
        for raster_path in raster_paths:
            src = rasterio.open(raster_path)
            sources.append(src)

        # Merge into a single mosaic array & transform
        arr, transform = rasterio.merge.merge(sources)

        # Copy metadata from the first dataset
        out_meta = sources[0].meta.copy()
        out_meta.update(
            {
                "driver": "GTiff",
                "height": arr.shape[1],
                "width": arr.shape[2],
                "transform": transform,
            }
        )
        cell_size = sources[0].res[0]
        if crs is None:
            crs = CRS.from_user_input(sources[0].crs)

        nodata = sources[0].nodata
        if nodata is not None:
            arr[arr == nodata] = np.nan

        arr = arr.squeeze().astype(np.float64)

        raster_meta = RasterMeta(cell_size=cell_size, crs=crs, transform=transform)
        raster_obj = Raster(arr=arr, raster_meta=raster_meta)
        return raster_obj
    finally:
        for src in sources:
            src.close()



"""Raster data structure."""

from __future__ import annotations

import importlib.util
import warnings
from collections.abc import Collection
from contextlib import contextmanager
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal, overload

import numpy as np
import numpy.ma
import rasterio.features
import rasterio.plot
import rasterio.sample
import rasterio.transform
import skimage.measure
from pydantic import BaseModel, InstanceOf, field_validator
from pyproj import Transformer
from pyproj.crs.crs import CRS
from rasterio.enums import Resampling
from rasterio.io import MemoryFile
from shapely.geometry import LineString, Point, Polygon

from rastr.arr.fill import fillna_nearest_neighbours
from rastr.gis.fishnet import create_fishnet
from rastr.gis.smooth import catmull_rom_smooth
from rastr.meta import RasterMeta

if TYPE_CHECKING:
    from collections.abc import Callable, Generator

    import geopandas as gpd
    from affine import Affine
    from branca.colormap import LinearColormap as BrancaLinearColormap
    from folium import Map
    from matplotlib.axes import Axes
    from matplotlib.image import AxesImage
    from numpy.typing import ArrayLike, NDArray
    from rasterio.io import BufferedDatasetWriter, DatasetReader, DatasetWriter
    from shapely import MultiPolygon
    from typing_extensions import Self

try:
    from rasterio._err import CPLE_BaseError
except ImportError:
    CPLE_BaseError = Exception  # Fallback if private module import fails


FOLIUM_INSTALLED = importlib.util.find_spec("folium") is not None
BRANCA_INSTALLED = importlib.util.find_spec("branca") is not None
MATPLOTLIB_INSTALLED = importlib.util.find_spec("matplotlib") is not None

CONTOUR_PERTURB_EPS = 1e-10


class RasterCellArrayShapeError(ValueError):
    """Custom error for invalid raster cell array shapes."""


class Raster(BaseModel):
    """2-dimensional raster and metadata."""

    arr: InstanceOf[np.ndarray]
    raster_meta: RasterMeta

    @field_validator("arr")
    @classmethod
    def check_2d_array(cls, v: NDArray) -> NDArray:
        """Validator to ensure the cell array is 2D."""
        if v.ndim != 2:
            msg = "Cell array must be 2D"
            raise RasterCellArrayShapeError(msg)
        return v

    @property
    def meta(self) -> RasterMeta:
        """Alias for raster_meta."""
        return self.raster_meta

    @meta.setter
    def meta(self, value: RasterMeta) -> None:
        self.raster_meta = value

    @property
    def shape(self) -> tuple[int, ...]:
        """Shape of the raster array."""
        return self.arr.shape

    @property
    def crs(self) -> CRS:
        """Convenience property to access the CRS via meta."""
        return self.meta.crs

    @crs.setter
    def crs(self, value: CRS) -> None:
        """Set the CRS via meta."""
        self.meta.crs = value

    @property
    def transform(self) -> Affine:
        """Convenience property to access the transform via meta."""
        return self.meta.transform

    @transform.setter
    def transform(self, value: Affine) -> None:
        """Set the transform via meta."""
        self.meta.transform = value

    def __init__(
        self,
        *,
        arr: ArrayLike,
        meta: RasterMeta | None = None,
        raster_meta: RasterMeta | None = None,
    ) -> None:
        arr = np.asarray(arr)

        # Set the meta
        if meta is not None and raster_meta is not None:
            msg = (
                "Only one of 'meta' or 'raster_meta' should be provided, they are "
                "aliases."
            )
            raise ValueError(msg)
        elif meta is not None and raster_meta is None:
            raster_meta = meta
        elif meta is None and raster_meta is not None:
            pass
        else:
            # Don't need to mention `'meta'` to simplify the messaging.
            msg = "The attribute 'raster_meta' is required."
            raise ValueError(msg)

        super().__init__(arr=arr, raster_meta=raster_meta)

    def __eq__(self, other: object) -> bool:
        """Check equality of two Raster objects."""
        if not isinstance(other, Raster):
            return NotImplemented
        return (
            np.array_equal(self.arr, other.arr)
            and self.raster_meta == other.raster_meta
        )

    def is_like(self, other: Raster) -> bool:
        """Check if two Raster objects have the same metadata and shape.

        Args:
            other: Another Raster to compare with.

        Returns:
            True if both rasters have the same meta and shape attributes.
        """
        return self.meta == other.meta and self.shape == other.shape

    __hash__ = BaseModel.__hash__

    def __add__(self, other: float | Self) -> Self:
        cls = self.__class__
        if isinstance(other, float | int):
            new_arr = self.arr + other
            return cls(arr=new_arr, raster_meta=self.raster_meta)
        elif isinstance(other, Raster):
            if self.raster_meta != other.raster_meta:
                msg = (
                    "Rasters must have the same metadata (e.g. CRS, cell size, etc.) "
                    "to be added"
                )
                raise ValueError(msg)
            if self.arr.shape != other.arr.shape:
                msg = (
                    "Rasters must have the same shape to be added:\n"
                    f"{self.arr.shape} != {other.arr.shape}"
                )
                raise ValueError(msg)
            new_arr = self.arr + other.arr
            return cls(arr=new_arr, raster_meta=self.raster_meta)
        else:
            return NotImplemented

    def __radd__(self, other: float) -> Self:
        return self + other

    def __mul__(self, other: float | Self) -> Self:
        cls = self.__class__
        if isinstance(other, float | int):
            new_arr = self.arr * other
            return cls(arr=new_arr, raster_meta=self.raster_meta)
        elif isinstance(other, Raster):
            if self.raster_meta != other.raster_meta:
                msg = (
                    "Rasters must have the same metadata (e.g. CRS, cell size, etc.) "
                    "to be multiplied"
                )
                raise ValueError(msg)
            if self.arr.shape != other.arr.shape:
                msg = "Rasters must have the same shape to be multiplied"
                raise ValueError(msg)
            new_arr = self.arr * other.arr
            return cls(arr=new_arr, raster_meta=self.raster_meta)
        else:
            return NotImplemented

    def __rmul__(self, other: float) -> Self:
        return self * other

    def __truediv__(self, other: float | Self) -> Self:
        cls = self.__class__
        if isinstance(other, float | int):
            new_arr = self.arr / other
            return cls(arr=new_arr, raster_meta=self.raster_meta)
        elif isinstance(other, Raster):
            if self.raster_meta != other.raster_meta:
                msg = (
                    "Rasters must have the same metadata (e.g. CRS, cell size, etc.) "
                    "to be divided"
                )
                raise ValueError(msg)
            if self.arr.shape != other.arr.shape:
                msg = "Rasters must have the same shape to be divided"
                raise ValueError(msg)
            new_arr = self.arr / other.arr
            return cls(arr=new_arr, raster_meta=self.raster_meta)
        else:
            return NotImplemented

    def __rtruediv__(self, other: float) -> Self:
        return self / other

    def __sub__(self, other: float | Self) -> Self:
        return self + (-other)

    def __rsub__(self, other: float) -> Self:
        return -self + other

    def __neg__(self) -> Self:
        cls = self.__class__
        return cls(arr=-self.arr, raster_meta=self.raster_meta)

    @property
    def cell_centre_coords(self) -> NDArray[np.float64]:
        """Get the coordinates of the cell centres in the raster."""
        return self.raster_meta.get_cell_centre_coords(self.arr.shape)

    @property
    def cell_x_coords(self) -> NDArray[np.float64]:
        """Get the x coordinates of the cell centres in the raster."""
        return self.raster_meta.get_cell_x_coords(self.arr.shape[1])

    @property
    def cell_y_coords(self) -> NDArray[np.float64]:
        """Get the y coordinates of the cell centres in the raster."""
        return self.raster_meta.get_cell_y_coords(self.arr.shape[0])

    @contextmanager
    def to_rasterio_dataset(
        self,
    ) -> Generator[DatasetReader | BufferedDatasetWriter | DatasetWriter]:
        """Create a rasterio in-memory dataset from the Raster object.

        Example:
            >>> raster = Raster.example()
            >>> with raster.to_rasterio_dataset() as dataset:
            >>>     ...
        """
        memfile = MemoryFile()

        height, width = self.arr.shape

        try:
            with memfile.open(
                driver="GTiff",
                height=height,
                width=width,
                count=1,  # Assuming a single band; adjust as necessary
                dtype=self.arr.dtype,
                crs=self.raster_meta.crs.to_wkt(),
                transform=self.raster_meta.transform,
            ) as dataset:
                dataset.write(self.arr, 1)

            # Yield the dataset for reading
            with memfile.open() as dataset:
                yield dataset
        finally:
            memfile.close()

    @overload
    def sample(
        self,
        xy: Collection[tuple[float, float]] | Collection[Point] | ArrayLike,
        *,
        na_action: Literal["raise", "ignore"] = "raise",
    ) -> NDArray: ...
    @overload
    def sample(
        self,
        xy: tuple[float, float] | Point,
        *,
        na_action: Literal["raise", "ignore"] = "raise",
    ) -> float: ...
    def sample(
        self,
        xy: Collection[tuple[float, float]]
        | Collection[Point]
        | ArrayLike
        | tuple[float, float]
        | Point,
        *,
        na_action: Literal["raise", "ignore"] = "raise",
    ) -> NDArray | float:
        """Sample raster values at GeoSeries locations and return sampled values.

        Args:
            xy: A list of (x, y) coordinates or shapely Point objects to sample the
                raster at.
            na_action: Action to take when a NaN value is encountered in the input xy.
                       Options are "raise" (raise an error) or "ignore" (replace with
                       NaN).

        Returns:
            A list of sampled raster values for each geometry in the GeoSeries.
        """
        # If this function is too slow, consider the optimizations detailed here:
        # https://rdrn.me/optimising-sampling/

        # Convert shapely Points to coordinate tuples if needed
        if isinstance(xy, Point):
            xy = [(xy.x, xy.y)]
            singleton = True
        elif (
            isinstance(xy, Collection)
            and len(xy) > 0
            and isinstance(next(iter(xy)), Point)
        ):
            xy = [(point.x, point.y) for point in xy]  # pyright: ignore[reportAttributeAccessIssue]
            singleton = False
        elif (
            isinstance(xy, tuple)
            and len(xy) == 2
            and isinstance(next(iter(xy)), (float, int))
        ):
            xy = [xy]  # pyright: ignore[reportAssignmentType]
            singleton = True
        else:
            singleton = False

        xy = np.asarray(xy, dtype=float)

        if len(xy) == 0:
            # Short-circuit
            return np.array([], dtype=float)

        # Create in-memory rasterio dataset from the incumbent Raster object
        with self.to_rasterio_dataset() as dataset:
            if dataset.count != 1:
                msg = "Only single band rasters are supported."
                raise NotImplementedError(msg)

            xy_arr = np.array(xy)

            # Determine the indexes of any x,y coordinates where either is NaN.
            # We will drop these indexes for the purposes of calling .sample, but
            # then we will add NaN values back in at the end, inserting NaN into the
            # results array.
            xy_is_nan = np.isnan(xy_arr).any(axis=1)
            xy_nan_idxs = list(np.atleast_1d(np.squeeze(np.nonzero(xy_is_nan))))
            xy_arr = xy_arr[~xy_is_nan]

            if na_action == "raise" and len(xy_nan_idxs) > 0:
                nan_error_msg = "NaN value found in input coordinates"
                raise ValueError(nan_error_msg)

            # Sample the raster in-memory dataset (e.g. PGA values) at the coordinates
            samples = list(
                rasterio.sample.sample_gen(
                    dataset,
                    xy_arr,
                    indexes=1,  # Single band raster, N.B. rasterio is 1-indexed
                    masked=True,
                )
            )

            # Convert the sampled values to a NumPy array and set masked values to NaN
            raster_values = np.array(
                [s.data[0] if not numpy.ma.getmask(s) else np.nan for s in samples]
            ).astype(float)

            if len(xy_nan_idxs) > 0:
                # Insert NaN values back into the results array
                # This is tricky because all the indexes get offset once we remove
                # elements.
                offset_xy_nan_idxs = xy_nan_idxs - np.arange(len(xy_nan_idxs))
                raster_values = np.insert(
                    raster_values,
                    offset_xy_nan_idxs,
                    np.nan,
                    axis=0,
                )

        if singleton:
            (raster_value,) = raster_values
            return raster_value

        return raster_values

    @property
    def bounds(self) -> tuple[float, float, float, float]:
        """Bounding box of the raster as (xmin, ymin, xmax, ymax)"""
        x1, y1, x2, y2 = rasterio.transform.array_bounds(
            height=self.arr.shape[0],
            width=self.arr.shape[1],
            transform=self.raster_meta.transform,
        )
        xmin, xmax = sorted([x1, x2])
        ymin, ymax = sorted([y1, y2])
        return (xmin, ymin, xmax, ymax)

    @property
    def bbox(self) -> Polygon:
        """Bounding box of the raster as a shapely polygon."""
        xmin, ymin, xmax, ymax = self.bounds
        return Polygon(
            [
                (xmin, ymin),
                (xmin, ymax),
                (xmax, ymax),
                (xmax, ymin),
                (xmin, ymin),
            ]
        )

    def explore(  # noqa: PLR0913 c.f. geopandas.explore which also has many input args
        self,
        *,
        m: Map | None = None,
        opacity: float = 1.0,
        colormap: str
        | Callable[[float], tuple[float, float, float, float]] = "viridis",
        cbar_label: str | None = None,
        vmin: float | None = None,
        vmax: float | None = None,
    ) -> Map:
        """Display the raster on a folium map."""
        if not FOLIUM_INSTALLED or not MATPLOTLIB_INSTALLED:
            msg = "The 'folium' and 'matplotlib' packages are required for 'explore()'."
            raise ImportError(msg)

        import folium.raster_layers
        import matplotlib as mpl

        if m is None:
            m = folium.Map()

        if vmin is not None and vmax is not None and vmax <= vmin:
            msg = "'vmin' must be less than 'vmax'."
            raise ValueError(msg)

        if isinstance(colormap, str):
            colormap = mpl.colormaps[colormap]

        # Transform bounds to WGS84 using pyproj directly
        wgs84_crs = CRS.from_epsg(4326)
        transformer = Transformer.from_crs(
            self.raster_meta.crs, wgs84_crs, always_xy=True
        )

        # Get the corner points of the bounding box
        raster_xmin, raster_ymin, raster_xmax, raster_ymax = self.bounds
        corner_points = [
            (raster_xmin, raster_ymin),
            (raster_xmin, raster_ymax),
            (raster_xmax, raster_ymax),
            (raster_xmax, raster_ymin),
        ]

        # Transform all corner points to WGS84
        transformed_points = [transformer.transform(x, y) for x, y in corner_points]

        # Find the bounding box of the transformed points
        transformed_xs, transformed_ys = zip(*transformed_points, strict=True)
        xmin, xmax = min(transformed_xs), max(transformed_xs)
        ymin, ymax = min(transformed_ys), max(transformed_ys)

        # Normalize the array to [0, 1] for colormap mapping
        _vmin, _vmax = _get_vmin_vmax(self, vmin=vmin, vmax=vmax)
        arr = self.normalize(vmin=_vmin, vmax=_vmax).arr

        # Finally, need to determine whether to flip the image based on negative Affine
        # coefficients
        flip_x = self.raster_meta.transform.a < 0
        flip_y = self.raster_meta.transform.e > 0
        if flip_x:
            arr = np.flip(arr, axis=1)
        if flip_y:
            arr = np.flip(arr, axis=0)

        bounds = [[ymin, xmin], [ymax, xmax]]
        img = folium.raster_layers.ImageOverlay(
            image=arr,
            bounds=bounds,
            opacity=opacity,
            colormap=colormap,
            mercator_project=True,
        )

        img.add_to(m)

        # Add a colorbar legend
        if BRANCA_INSTALLED:
            cbar = _map_colorbar(colormap=colormap, vmin=_vmin, vmax=_vmax)
            if cbar_label:
                cbar.caption = cbar_label
            cbar.add_to(m)

        m.fit_bounds(bounds)

        return m

    def normalize(
        self, *, vmin: float | None = None, vmax: float | None = None
    ) -> Self:
        """Normalize the raster values to the range [0, 1].

        If custom vmin and vmax values are provided, values below vmin will be set to 0,
        and values above vmax will be set to 1.

        Args:
            vmin: Minimum value for normalization. Values below this will be set to 0.
                  If None, the minimum value in the array is used.
            vmax: Maximum value for normalization. Values above this will be set to 1.
                  If None, the maximum value in the array is used.
        """
        _vmin, _vmax = _get_vmin_vmax(self, vmin=vmin, vmax=vmax)

        arr = self.arr.copy()
        if _vmax > _vmin:
            arr = (arr - _vmin) / (_vmax - _vmin)
            arr = np.clip(arr, 0, 1)
        else:
            arr = np.zeros_like(arr)
        return self.__class__(arr=arr, raster_meta=self.raster_meta)

    def to_clipboard(self) -> None:
        """Copy the raster cell array to the clipboard."""
        import pandas as pd

        pd.DataFrame(self.arr).to_clipboard(index=False, header=False)

    def plot(
        self,
        *,
        ax: Axes | None = None,
        cbar_label: str | None = None,
        basemap: bool = False,
        cmap: str = "viridis",
        suppressed: Collection[float] | float = tuple(),
        **kwargs: Any,
    ) -> Axes:
        """Plot the raster on a matplotlib axis.

        Args:
            ax: A matplotlib axes object to plot on. If None, a new figure will be
                created.
            cbar_label: Label for the colorbar. If None, no label is added.
            basemap: Whether to add a basemap. Currently not implemented.
            cmap: Colormap to use for the plot.
            suppressed: Values to suppress from the plot (i.e. not display). This can be
                        useful for zeroes especially.
            **kwargs: Additional keyword arguments to pass to `rasterio.plot.show()`.
                      This includes parameters like `alpha` for transparency.
        """
        if not MATPLOTLIB_INSTALLED:
            msg = "The 'matplotlib' package is required for 'plot()'."
            raise ImportError(msg)

        from matplotlib import pyplot as plt
        from mpl_toolkits.axes_grid1 import make_axes_locatable

        suppressed = np.array(suppressed)

        if ax is None:
            _, _ax = plt.subplots()
            _ax: Axes
            ax = _ax

        if basemap:
            msg = "Basemap plotting is not yet implemented."
            raise NotImplementedError(msg)

        model = self.model_copy()
        model.arr = model.arr.copy()

        # Get extent of the unsuppressed values in array index coordinates
        suppressed_mask = np.isin(model.arr, suppressed)
        (x_unsuppressed,) = np.nonzero((~suppressed_mask).any(axis=0))
        (y_unsuppressed,) = np.nonzero((~suppressed_mask).any(axis=1))

        if len(x_unsuppressed) == 0 or len(y_unsuppressed) == 0:
            msg = "Raster contains no unsuppressed values; cannot plot."
            raise ValueError(msg)

        # N.B. these are array index coordinates, so np.min and np.max are safe since
        # they cannot encounter NaN values.
        min_x_unsuppressed = np.min(x_unsuppressed)
        max_x_unsuppressed = np.max(x_unsuppressed)
        min_y_unsuppressed = np.min(y_unsuppressed)
        max_y_unsuppressed = np.max(y_unsuppressed)

        # Transform to raster CRS
        x1, y1 = self.raster_meta.transform * (min_x_unsuppressed, min_y_unsuppressed)  # type: ignore[reportAssignmentType] overloaded tuple size in affine
        x2, y2 = self.raster_meta.transform * (max_x_unsuppressed, max_y_unsuppressed)  # type: ignore[reportAssignmentType]
        xmin, xmax = sorted([x1, x2])
        ymin, ymax = sorted([y1, y2])

        model.arr[suppressed_mask] = np.nan

        img, *_ = model.rio_show(ax=ax, cmap=cmap, with_bounds=True, **kwargs)

        ax.set_xlim(xmin, xmax)
        ax.set_ylim(ymin, ymax)

        ax.set_aspect("equal", "box")
        ax.set_yticklabels([])
        ax.set_xticklabels([])

        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        fig = ax.get_figure()
        if fig is not None:
            fig.colorbar(img, label=cbar_label, cax=cax)
        return ax

    def rio_show(self, **kwargs: Any) -> list[AxesImage]:
        """Plot the raster using rasterio's built-in plotting function.

        This is useful for lower-level access to rasterio's plotting capabilities.
        Generally, the `plot()` method is preferred for most use cases.

        Args:
            **kwargs: Keyword arguments to pass to `rasterio.plot.show()`. This includes
            parameters like `alpha` for transparency, and `with_bounds` to control
            whether to plot in spatial coordinates or array index coordinates.
        """
        with self.to_rasterio_dataset() as dataset:
            return rasterio.plot.show(dataset, **kwargs).get_images()

    def as_geodataframe(self, name: str = "value") -> gpd.GeoDataFrame:
        """Create a GeoDataFrame representation of the raster."""
        import geopandas as gpd

        polygons = create_fishnet(bounds=self.bounds, res=self.raster_meta.cell_size)
        point_tuples = [polygon.centroid.coords[0] for polygon in polygons]
        raster_gdf = gpd.GeoDataFrame(
            {
                "geometry": polygons,
                name: self.sample(point_tuples, na_action="ignore"),
            },
            crs=self.raster_meta.crs,
        )

        return raster_gdf

    def to_file(self, path: Path | str) -> None:
        """Write the raster to a GeoTIFF file."""

        path = Path(path)

        suffix = path.suffix.lower()
        if suffix in (".tif", ".tiff"):
            driver = "GTiff"
        elif suffix in (".grd"):
            # https://grapherhelp.goldensoftware.com/subsys/ascii_grid_file_format.htm
            # e.g. Used by AnAqSim
            driver = "GSAG"
        else:
            msg = f"Unsupported file extension: {suffix}"
            raise ValueError(msg)

        with rasterio.open(
            path,
            "w",
            driver=driver,
            height=self.arr.shape[0],
            width=self.arr.shape[1],
            count=1,
            dtype=self.arr.dtype,
            crs=self.raster_meta.crs,
            transform=self.raster_meta.transform,
            nodata=np.nan,
        ) as dst:
            try:
                dst.write(self.arr, 1)
            except CPLE_BaseError as err:
                msg = f"Failed to write raster to file: {err}"
                raise OSError(msg) from err

    def __str__(self) -> str:
        cls = self.__class__
        mean = np.nanmean(self.arr)
        return f"{cls.__name__}(shape={self.arr.shape}, {mean=})"

    def __repr__(self) -> str:
        return str(self)

    @classmethod
    def example(cls) -> Self:
        """Create an example Raster."""
        # Peaks dataset style example
        n = 256
        x = np.linspace(-3, 3, n)
        y = np.linspace(-3, 3, n)
        x, y = np.meshgrid(x, y)
        z = np.exp(-(x**2) - y**2) * np.sin(3 * np.sqrt(x**2 + y**2))
        arr = z.astype(np.float32)

        raster_meta = RasterMeta.example()
        return cls(arr=arr, raster_meta=raster_meta)

    @overload
    def apply(
        self,
        func: Callable[[np.ndarray], np.ndarray],
        *,
        raw: Literal[True],
    ) -> Self: ...
    @overload
    def apply(
        self,
        func: Callable[[float], float] | Callable[[np.ndarray], np.ndarray],
        *,
        raw: Literal[False] = False,
    ) -> Self: ...
    def apply(self, func, *, raw=False) -> Self:
        """Apply a function element-wise to the raster array.

        Creates a new raster instance with the same metadata (CRS, transform, etc.)
        but with the data array transformed by the provided function. The original
        raster is not modified.

        Args:
            func: The function to apply to the raster array. If `raw` is True, this
                  function should accept and return a NumPy array. If `raw` is False,
                  this function should accept and return a single float value.
            raw: If True, the function is applied directly to the entire array at
                 once. If False, the function is applied element-wise to each cell
                 in the array using `np.vectorize()`. Default is False.
        """
        new_raster = self.model_copy()
        if raw:
            new_arr = func(self.arr)
        else:
            new_arr = np.vectorize(func)(self.arr)
        new_raster.arr = np.asarray(new_arr)
        return new_raster

    def max(self) -> float:
        """Get the maximum value in the raster, ignoring NaN values.

        Returns:
            The maximum value in the raster. Returns NaN if all values are NaN.
        """
        return float(np.nanmax(self.arr))

    def min(self) -> float:
        """Get the minimum value in the raster, ignoring NaN values.

        Returns:
            The minimum value in the raster. Returns NaN if all values are NaN.
        """
        return float(np.nanmin(self.arr))

    def mean(self) -> float:
        """Get the mean value in the raster, ignoring NaN values.

        Returns:
            The mean value in the raster. Returns NaN if all values are NaN.
        """
        return float(np.nanmean(self.arr))

    def std(self) -> float:
        """Get the standard deviation of values in the raster, ignoring NaN values.

        Returns:
            The standard deviation of the raster. Returns NaN if all values are NaN.
        """
        return float(np.nanstd(self.arr))

    def quantile(self, q: float) -> float:
        """Get the specified quantile value in the raster, ignoring NaN values.

        Args:
            q: Quantile to compute, must be between 0 and 1 inclusive.

        Returns:
            The quantile value. Returns NaN if all values are NaN.
        """
        return float(np.nanquantile(self.arr, q))

    def median(self) -> float:
        """Get the median value in the raster, ignoring NaN values.

        This is equivalent to quantile(0.5).

        Returns:
            The median value in the raster. Returns NaN if all values are NaN.
        """
        return float(np.nanmedian(self.arr))

    def fillna(self, value: float) -> Self:
        """Fill NaN values in the raster with a specified value.

        See also `extrapolate()` for filling NaN values using extrapolation from data.
        """
        filled_arr = np.nan_to_num(self.arr, nan=value)
        new_raster = self.model_copy()
        new_raster.arr = filled_arr
        return new_raster

    def get_xy(self) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        """Get the x and y coordinates of the raster cell centres in meshgrid format.

        Returns the coordinates of the cell centres as two separate 2D arrays in
        meshgrid format, where each array has the same shape as the raster data array.

        Returns:
            A tuple of (x, y) coordinate arrays where:
            - x: 2D array of x-coordinates of cell centres
            - y: 2D array of y-coordinates of cell centres
            Both arrays have the same shape as the raster data array.
        """
        coords = self.raster_meta.get_cell_centre_coords(self.arr.shape)
        return coords[:, :, 0], coords[:, :, 1]

    def contour(
        self, levels: list[float] | NDArray, *, smoothing: bool = True
    ) -> gpd.GeoDataFrame:
        """Create contour lines from the raster data, optionally with smoothing.

        The contour lines are returned as a GeoDataFrame with the contours dissolved
        by level, resulting in one row per contour level. Each row contains a
        (Multi)LineString geometry representing all contour lines for that level,
        and the contour level value in a column named 'level'.

        Consider calling `blur()` before this method to smooth the raster data before
        contouring, to denoise the contours.

        Args:
            levels: A list or array of contour levels to generate. The contour lines
                    will be generated for each level in this sequence.
            smoothing: Defaults to true, which corresponds to applying a smoothing
                       algorithm to the contour lines. At the moment, this is the
                       Catmull-Rom spline algorithm. If set to False, the raw
                       contours will be returned without any smoothing.
        """
        import geopandas as gpd

        all_levels = []
        all_geoms = []
        for level in levels:
            # If this is the maximum or minimum level, perturb it ever-so-slightly to
            # ensure we get contours at the edges of the raster
            perturbed_level = level
            if level == self.max():
                perturbed_level -= CONTOUR_PERTURB_EPS
            elif level == self.min():
                perturbed_level += CONTOUR_PERTURB_EPS

            contours = skimage.measure.find_contours(
                self.arr,
                level=perturbed_level,
            )

            # Construct shapely LineString objects
            # Convert to CRS from array index coordinates to raster CRS
            geoms = [
                LineString(
                    np.array(
                        rasterio.transform.xy(self.raster_meta.transform, *contour.T)
                    ).T
                )
                for contour in contours
                # Contour lines need at least three distinct points to avoid
                # degenerate geometries
                if np.unique(contour, axis=0).shape[0] > 2
            ]

            # Apply smoothing if requested
            if smoothing:
                geoms = [catmull_rom_smooth(geom) for geom in geoms]

            all_geoms.extend(geoms)
            all_levels.extend([level] * len(geoms))

        contour_gdf = gpd.GeoDataFrame(
            data={
                "level": all_levels,
            },
            geometry=all_geoms,
            crs=self.raster_meta.crs,
        )

        # Dissolve contours by level to merge all contour lines of the same level
        return contour_gdf.dissolve(by="level", as_index=False)

    def blur(self, sigma: float) -> Self:
        """Apply a Gaussian blur to the raster data.

        Args:
            sigma: Standard deviation for Gaussian kernel, in units of geographic
                   coordinate distance (e.g. meters). A larger sigma results in a more
                   blurred image.
        """
        from scipy.ndimage import gaussian_filter

        cell_sigma = sigma / self.raster_meta.cell_size

        blurred_array = gaussian_filter(self.arr, sigma=cell_sigma)
        new_raster = self.model_copy()
        new_raster.arr = blurred_array
        return new_raster

    def extrapolate(self, method: Literal["nearest"] = "nearest") -> Self:
        """Extrapolate the raster data to fill NaN values.

        See also `fillna()` for filling NaN values with a specific value.

        If the raster is all-NaN, this method will return a copy of the raster without
        changing the NaN values.

        Args:
            method: The method to use for extrapolation. Currently only 'nearest' is
                    supported, which fills NaN values with the nearest non-NaN value.
        """
        if method not in ("nearest",):
            msg = f"Unsupported extrapolation method: {method}"
            raise NotImplementedError(msg)

        raster = self.model_copy()
        raster.arr = fillna_nearest_neighbours(arr=self.arr)

        return raster

    def pad(self, width: float, *, value: float = np.nan) -> Self:
        """Extend the raster by adding a constant fill value around the edges.

        By default, the padding value is NaN, but this can be changed via the
        `value` parameter.

        This grows the raster by adding padding around all edges. New cells are
        filled with the constant `value`.

        If the width is not an exact multiple of the cell size, the padding may be
        slightly larger than the specified width, i.e. the value is rounded up to
        the nearest whole number of cells.

        Args:
            width: The width of the padding, in the same units as the raster CRS
                   (e.g. meters). This defines how far from the edge the padding
                   extends.
            value: The constant value to use for padding. Default is NaN.
        """
        cell_size = self.raster_meta.cell_size

        # Calculate number of cells to pad in each direction
        pad_cells = int(np.ceil(width / cell_size))

        # Get current bounds
        xmin, ymin, xmax, ymax = self.bounds

        # Calculate new bounds with padding
        new_xmin = xmin - (pad_cells * cell_size)
        new_ymin = ymin - (pad_cells * cell_size)
        new_xmax = xmax + (pad_cells * cell_size)
        new_ymax = ymax + (pad_cells * cell_size)

        # Create padded array
        new_height = self.arr.shape[0] + 2 * pad_cells
        new_width = self.arr.shape[1] + 2 * pad_cells

        # Create new array filled with the padding value
        padded_arr = np.full((new_height, new_width), value, dtype=self.arr.dtype)

        # Copy original array into the center of the padded array
        padded_arr[
            pad_cells : pad_cells + self.arr.shape[0],
            pad_cells : pad_cells + self.arr.shape[1],
        ] = self.arr

        # Create new transform for the padded raster
        new_transform = rasterio.transform.from_bounds(
            west=new_xmin,
            south=new_ymin,
            east=new_xmax,
            north=new_ymax,
            width=new_width,
            height=new_height,
        )

        # Create new raster metadata
        new_meta = RasterMeta(
            cell_size=cell_size,
            crs=self.raster_meta.crs,
            transform=new_transform,
        )

        return self.__class__(arr=padded_arr, raster_meta=new_meta)

    def crop(
        self,
        bounds: tuple[float, float, float, float],
        *,
        strategy: Literal["underflow", "overflow"] = "underflow",
    ) -> Self:
        """Crop the raster to the specified bounds as (minx, miny, maxx, maxy).

        Args:
            bounds: A tuple of (minx, miny, maxx, maxy) defining the bounds to crop to.
            strategy:   The cropping strategy to use. 'underflow' will crop the raster
                        to be fully within the bounds, ignoring any cells that are
                        partially outside the bounds. 'overflow' will instead include
                        cells that intersect the bounds, ensuring the bounds area
                        remains covered with cells.

        Returns:
            A new Raster instance cropped to the specified bounds.
        """

        minx, miny, maxx, maxy = bounds
        arr = self.arr

        # Get the half cell size for cropping
        cell_size = self.raster_meta.cell_size
        half_cell_size = cell_size / 2

        # Get the cell centre coordinates as 1D arrays
        x_coords = self.cell_x_coords
        y_coords = self.cell_y_coords

        # Get the indices to crop the array
        if strategy == "underflow":
            x_idx = (x_coords >= minx + half_cell_size) & (
                x_coords <= maxx - half_cell_size
            )
            y_idx = (y_coords >= miny + half_cell_size) & (
                y_coords <= maxy - half_cell_size
            )
        elif strategy == "overflow":
            x_idx = (x_coords > minx - half_cell_size) & (
                x_coords < maxx + half_cell_size
            )
            y_idx = (y_coords > miny - half_cell_size) & (
                y_coords < maxy + half_cell_size
            )
        else:
            msg = f"Unsupported cropping strategy: {strategy}"
            raise NotImplementedError(msg)

        # Crop the array
        cropped_arr = arr[np.ix_(y_idx, x_idx)]

        # Check the shape of the cropped array
        if cropped_arr.size == 0:
            msg = "Cropped array is empty; no cells within the specified bounds."
            raise ValueError(msg)

        # Recalculate the transform for the cropped raster
        x_coords = x_coords[x_idx]
        y_coords = y_coords[y_idx]
        transform = rasterio.transform.from_bounds(
            west=x_coords.min() - half_cell_size,
            south=y_coords.min() - half_cell_size,
            east=x_coords.max() + half_cell_size,
            north=y_coords.max() + half_cell_size,
            width=cropped_arr.shape[1],
            height=cropped_arr.shape[0],
        )

        # Update the raster
        cls = self.__class__
        new_meta = RasterMeta(
            cell_size=cell_size, crs=self.raster_meta.crs, transform=transform
        )
        return cls(arr=cropped_arr, raster_meta=new_meta)

    def taper_border(self, width: float, *, limit: float = 0.0) -> Self:
        """Taper values to a limiting value around the border of the raster.

        By default, the borders are tapered to zero, but this can be changed via the
        `limit` parameter.

        This keeps the raster size the same, overwriting values in the border area.
        To instead grow the raster, consider using `pad()` followed by `taper_border()`.

        The tapering is linear from the cell centres around the border of the raster,
        so the value at the edge of the raster will be equal to `limit`.

        Args:
            width: The width of the taper, in the same units as the raster CRS
                   (e.g. meters). This defines how far from the edge the tapering
                   starts.
            limit: The limiting value to taper to at the edges. Default is zero.
        """

        # Determine the width in cell units (possibly fractional)
        cell_size = self.raster_meta.cell_size
        width_in_cells = width / cell_size

        # Calculate the distance from the edge in cell units
        arr_height, arr_width = self.arr.shape
        y_indices, x_indices = np.indices((int(arr_height), int(arr_width)))
        dist_from_left = x_indices
        dist_from_right = arr_width - 1 - x_indices
        dist_from_top = y_indices
        dist_from_bottom = arr_height - 1 - y_indices
        dist_from_edge = np.minimum.reduce(
            [dist_from_left, dist_from_right, dist_from_top, dist_from_bottom]
        )

        # Mask the arrays to only the area within the width from the edge, rounding up
        mask = dist_from_edge < np.ceil(width_in_cells)
        masked_dist_arr = np.where(mask, dist_from_edge, np.nan)
        masked_arr = np.where(mask, self.arr, np.nan)

        # Calculate the tapering factor based on the distance from the edge
        taper_factor = np.clip(masked_dist_arr / width_in_cells, 0.0, 1.0)
        tapered_values = limit + (masked_arr - limit) * taper_factor

        # Create the new raster array
        new_arr = self.arr.copy()
        new_arr[mask] = tapered_values[mask]
        new_raster = self.model_copy()
        new_raster.arr = new_arr

        return new_raster

    def clip(
        self,
        polygon: Polygon | MultiPolygon,
        *,
        strategy: Literal["centres"] = "centres",
    ) -> Self:
        """Clip the raster to the specified polygon, replacing cells outside with NaN.

        The clipping strategy determines how to handle cells that are partially
        within the polygon. Currently, only the 'centres' strategy is supported, which
        retains cells whose centres fall within the polygon.

        Args:
            polygon: A shapely Polygon or MultiPolygon defining the area to clip to.
            strategy: The clipping strategy to use. Currently only 'centres' is
                      supported, which retains cells whose centres fall within the
                      polygon.

        Returns:
            A new Raster with cells outside the polygon set to NaN.
        """
        if strategy != "centres":
            msg = f"Unsupported clipping strategy: {strategy}"
            raise NotImplementedError(msg)

        raster = self.model_copy()

        mask = rasterio.features.rasterize(
            [(polygon, 1)],
            fill=0,
            out_shape=self.shape,
            transform=self.meta.transform,
            dtype=np.uint8,
        )

        raster.arr = np.where(mask, raster.arr, np.nan)

        return raster

    def trim_nan(self) -> Self:
        """Crop the raster by trimming away all-NaN slices at the edges.

        This effectively trims the raster to the smallest bounding box that contains all
        of the non-NaN values. Note that this does not guarantee no NaN values at all
        around the edges, only that there won't be entire edges which are all-NaN.

        Consider using `.extrapolate()` for further cleanup of NaN values.
        """
        arr = self.arr

        # Check if the entire array is NaN
        if np.all(np.isnan(arr)):
            msg = "Cannot crop raster: all values are NaN"
            raise ValueError(msg)

        # Find rows and columns that are not all NaN
        nan_row_mask = np.all(np.isnan(arr), axis=1)
        nan_col_mask = np.all(np.isnan(arr), axis=0)

        # Find the bounding indices
        (row_indices,) = np.where(~nan_row_mask)
        (col_indices,) = np.where(~nan_col_mask)

        min_row, max_row = row_indices[0], row_indices[-1]
        min_col, max_col = col_indices[0], col_indices[-1]

        # Crop the array
        cropped_arr = arr[min_row : max_row + 1, min_col : max_col + 1]

        # Shift the transform by the number of pixels cropped (min_col, min_row)
        new_transform = (
            self.raster_meta.transform
            * rasterio.transform.Affine.translation(min_col, min_row)
        )

        # Create new metadata
        new_meta = RasterMeta(
            cell_size=self.raster_meta.cell_size,
            crs=self.raster_meta.crs,
            transform=new_transform,
        )

        return self.__class__(arr=cropped_arr, raster_meta=new_meta)

    def resample(
        self, new_cell_size: float, *, method: Literal["bilinear"] = "bilinear"
    ) -> Self:
        """Resample the raster data to a new resolution.

        If the new cell size is not an exact multiple of the current cell size, the
        overall raster bounds may increase slightly. The affine transform will keep
        the same shift, i.e. the top-left corner of the raster will remain in the same'
        coordinate location. A corollary is that the overall centre of the raster bounds
        will not necessary be the same as the original raster.

        Args:
            new_cell_size: The desired cell size for the resampled raster.
            method: The resampling method to use. Only 'bilinear' is supported.
        """
        if method not in ("bilinear",):
            msg = f"Unsupported resampling method: {method}"
            raise NotImplementedError(msg)

        factor = self.raster_meta.cell_size / new_cell_size

        cls = self.__class__
        # Use the rasterio dataset with proper context management
        with self.to_rasterio_dataset() as dataset:
            # N.B. the new height and width may increase slightly.
            new_height = int(np.ceil(dataset.height * factor))
            new_width = int(np.ceil(dataset.width * factor))

            # Resample via rasterio
            (new_arr,) = dataset.read(  # Assume exactly one band
                out_shape=(dataset.count, new_height, new_width),
                resampling=Resampling.bilinear,
            )

            # Create new RasterMeta with updated transform and cell size
            new_raster_meta = RasterMeta(
                transform=dataset.transform
                * dataset.transform.scale(
                    (dataset.width / new_width),
                    (dataset.height / new_height),
                ),
                crs=self.raster_meta.crs,
                cell_size=new_cell_size,
            )

            return cls(arr=new_arr, raster_meta=new_raster_meta)


def _map_colorbar(
    *,
    colormap: Callable[[float], tuple[float, float, float, float]],
    vmin: float,
    vmax: float,
) -> BrancaLinearColormap:
    from branca.colormap import LinearColormap as BrancaLinearColormap
    from matplotlib.colors import ListedColormap, to_hex

    # Determine legend data range in original units
    vmin = float(vmin) if np.isfinite(vmin) else 0.0
    vmax = float(vmax) if np.isfinite(vmax) else 1.0
    if vmax <= vmin:
        vmax = vmin + 1.0

    if isinstance(colormap, ListedColormap):
        n = colormap.N
    else:
        n = 256

    sample_points = np.linspace(0, 1, n)
    colors = [to_hex(colormap(x)) for x in sample_points]
    return BrancaLinearColormap(colors=colors, vmin=vmin, vmax=vmax)


def _get_vmin_vmax(
    raster: Raster, *, vmin: float | None = None, vmax: float | None = None
) -> tuple[float, float]:
    """Get maximum and minimum values from a raster array, ignoring NaNs.

    Allows for custom over-ride vmin and vmax values to be provided.
    """
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message="All-NaN slice encountered",
            category=RuntimeWarning,
        )
        if vmin is None:
            _vmin = raster.min()
        else:
            _vmin = vmin
        if vmax is None:
            _vmax = raster.max()
        else:
            _vmax = vmax

    return _vmin, _vmax


RasterModel = Raster



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import pytest
import rasterio.transform
from affine import Affine
from pyproj.crs.crs import CRS

from rastr.io import read_raster_inmem, read_raster_mosaic_inmem

if TYPE_CHECKING:
    from pathlib import Path

_WGS84_CRS = CRS.from_epsg(4326)


class TestReadRasterInMem:
    def test_small_tif(self, assets_dir: Path):
        raster_path = assets_dir / "pga_g_clipped.tif"
        raster_obj = read_raster_inmem(raster_path)

        assert raster_obj.arr.shape == (2, 2)
        assert raster_obj.raster_meta.cell_size == 0.01495024875620743
        assert raster_obj.raster_meta.crs.to_epsg() == 4326
        assert raster_obj.raster_meta.transform == Affine(
            0.01495024875620743,
            0.0,
            173.7895771144279,
            0.0,
            -0.014950248756214535,
            -41.82587064676616,
        )

        assert raster_obj.arr[0, 0] == pytest.approx(0.39016372)
        assert raster_obj.arr[1, 1] == pytest.approx(0.4411124)
        assert raster_obj.arr[0, 1] == pytest.approx(0.44069204)
        assert raster_obj.arr[1, 0] == pytest.approx(0.41911235)

    def test_small_grd(self, assets_dir: Path):
        raster_path = assets_dir / "pga_g_clipped.grd"
        raster_obj = read_raster_inmem(raster_path)

        assert raster_obj.arr.shape == (2, 2)
        assert raster_obj.raster_meta.cell_size == 0.01495024875620743
        assert raster_obj.raster_meta.crs.to_epsg() == 4326
        assert raster_obj.raster_meta.transform == Affine(
            0.01495024875620743,
            0.0,
            173.7895771144279,
            0.0,
            -0.014950248756214535,
            -41.82587064676616,
        )

        assert raster_obj.arr[0, 0] == pytest.approx(0.39016372)
        assert raster_obj.arr[1, 1] == pytest.approx(0.4411124)
        assert raster_obj.arr[0, 1] == pytest.approx(0.44069204)
        assert raster_obj.arr[1, 0] == pytest.approx(0.41911235)

    def test_str_crs(self, assets_dir: Path):
        # Arrange
        raster_path = assets_dir / "pga_g_clipped.tif"
        crs = "EPSG:4326"

        # Act
        raster_obj = read_raster_inmem(raster_path, crs=crs)

        # Assert
        assert raster_obj.raster_meta.crs.to_epsg() == 4326

    def test_pyproj_crs(self, assets_dir: Path):
        # Arrange
        raster_path = assets_dir / "pga_g_clipped.tif"
        crs = CRS.from_epsg(4326)

        # Act
        raster_obj = read_raster_inmem(raster_path, crs=crs)

        # Assert
        assert raster_obj.raster_meta.crs.to_epsg() == 4326


class TestReadRasterMosaicInMem:
    @staticmethod
    def _write_tiff(
        dir_path: Path,
        name: str,
        arr: np.ndarray,
        transform: rasterio.transform.Affine,
        crs: str | CRS = "EPSG:4326",
        nodata: float = float("nan"),
    ) -> Path:
        path = dir_path / name
        with rasterio.open(
            path,
            "w",
            driver="GTiff",
            height=arr.shape[0],
            width=arr.shape[1],
            count=1,
            dtype=arr.dtype,
            crs=crs,
            transform=transform,
            nodata=nodata,
        ) as dst:
            dst.write(arr, 1)
        return path

    def test_basic_and_nodata(self, tmp_path: Path):
        # Arrange
        tile1 = np.array([[1, 2], [3, -9999]], dtype=np.float32)
        tile2 = np.array([[5, 6], [7, 8]], dtype=np.float32)
        transform1 = rasterio.transform.Affine(1, 0, 0, 0, -1, 2)
        transform2 = rasterio.transform.Affine(1, 0, 2, 0, -1, 2)
        self._write_tiff(tmp_path, "tile1.tif", tile1, transform1, nodata=-9999)
        self._write_tiff(tmp_path, "tile2.tif", tile2, transform2)

        # Act
        raster = read_raster_mosaic_inmem(tmp_path)

        # Assert
        assert raster.arr.shape == (2, 4)
        assert raster.raster_meta.transform == transform1
        assert raster.raster_meta.cell_size == 1
        assert np.isnan(raster.arr[1, 1])
        expected = np.array([[1, 2, 5, 6], [3, np.nan, 7, 8]], dtype=np.float64)
        np.testing.assert_allclose(raster.arr, expected, equal_nan=True)

    def test_custom_glob(self, tmp_path: Path):
        # Arrange
        arr_a = np.array([[10, 11], [12, 13]], dtype=np.float32)
        arr_b = np.array([[20, 21], [22, 23]], dtype=np.float32)
        transform_a = rasterio.transform.Affine(1, 0, 0, 0, -1, 2)
        transform_b = rasterio.transform.Affine(1, 0, 2, 0, -1, 2)
        self._write_tiff(tmp_path, "keep_a.tif", arr_a, transform_a)
        self._write_tiff(tmp_path, "skip_b.tif", arr_b, transform_b)

        # Act
        raster = read_raster_mosaic_inmem(tmp_path, glob="keep_*.tif")

        # Assert
        assert raster.arr.shape == (2, 2)
        np.testing.assert_allclose(raster.arr, arr_a)

    def test_no_files(self, tmp_path: Path):
        # Arrange / Act / Assert
        with pytest.raises(FileNotFoundError):
            read_raster_mosaic_inmem(tmp_path)

    def test_override_crs(self, tmp_path: Path):
        # Arrange
        arr = np.array([[1, 2], [3, 4]], dtype=np.float32)
        transform = rasterio.transform.Affine(1, 0, 0, 0, -1, 2)
        self._write_tiff(tmp_path, "tile.tif", arr, transform)
        override_crs = CRS.from_epsg(3857)

        # Act
        raster = read_raster_mosaic_inmem(tmp_path, crs=override_crs)

        # Assert
        assert raster.raster_meta.crs.to_epsg() == 3857
        np.testing.assert_allclose(raster.arr, arr.astype(np.float64))

    def test_vertical_adjacency(self, tmp_path: Path):
        # Arrange
        top = np.array([[1, 2], [3, 4]], dtype=np.float32)
        bottom = np.array([[5, 6], [7, 8]], dtype=np.float32)
        transform_top = rasterio.transform.Affine(1, 0, 0, 0, -1, 4)
        transform_bottom = rasterio.transform.Affine(1, 0, 0, 0, -1, 2)
        self._write_tiff(tmp_path, "top.tif", top, transform_top)
        self._write_tiff(tmp_path, "bottom.tif", bottom, transform_bottom)

        # Act
        raster = read_raster_mosaic_inmem(tmp_path)

        # Assert
        assert raster.arr.shape == (4, 2)
        assert raster.raster_meta.transform == transform_top
        expected = np.vstack([top, bottom]).astype(np.float64)
        np.testing.assert_allclose(raster.arr, expected)

    def test_overlapping_first_wins(self, tmp_path: Path):
        # Arrange
        first = np.array([[1, 1], [1, 1]], dtype=np.float32)
        second = np.array([[2, 2], [2, 2]], dtype=np.float32)
        transform = rasterio.transform.Affine(1, 0, 0, 0, -1, 2)
        # File naming ensures "a_" is encountered before "z_" in glob listing
        self._write_tiff(tmp_path, "a_first.tif", first, transform)
        self._write_tiff(tmp_path, "z_second.tif", second, transform)

        # Act
        raster = read_raster_mosaic_inmem(tmp_path)

        # Assert
        np.testing.assert_allclose(raster.arr, first.astype(np.float64))



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from affine import Affine
from pydantic import BaseModel, InstanceOf
from pyproj import CRS

if TYPE_CHECKING:
    from numpy.typing import NDArray
    from typing_extensions import Self


class RasterMeta(BaseModel, extra="forbid"):
    """Raster metadata.

    Attributes:
        cell_size: Cell size in meters.
        crs: Coordinate reference system.
        transform: The affine transformation associated with the raster. This is based
                   on the CRS, the cell size, as well as the offset/origin.
    """

    cell_size: float
    crs: InstanceOf[CRS]
    transform: InstanceOf[Affine]

    @classmethod
    def example(cls) -> Self:
        """Create an example RasterMeta object."""
        return cls(
            cell_size=2.0,
            crs=CRS.from_epsg(2193),
            transform=Affine.scale(2.0, 2.0),
        )

    def get_cell_centre_coords(self, shape: tuple[int, int]) -> NDArray:
        """Return an array of (x, y) coordinates for the center of each cell.

        The coordinates will be in the coordinate system defined by the
        raster's transform.

        Args:
            shape: (rows, cols) of the raster array.

        Returns:
            (x, y) coordinates for each cell center, with shape (rows, cols, 2)
        """
        x_coords = self.get_cell_x_coords(shape[1])  # cols for x-coordinates
        y_coords = self.get_cell_y_coords(shape[0])  # rows for y-coordinates
        coords = np.stack(np.meshgrid(x_coords, y_coords), axis=-1)
        return coords

    def get_cell_x_coords(self, n_columns: int) -> NDArray:
        """Return an array of x coordinates for the center of each cell.

        The coordinates will be in the coordinate system defined by the
        raster's transform.

        Args:
            n_columns: Number of columns in the raster array.

        Returns:
            x_coordinates at cell centers, with shape (n_columns,)
        """
        x_idx = np.arange(n_columns) + 0.5
        y_idx = np.zeros_like(x_idx)  # Use y=0 for a single row
        x_coords, _ = self.transform * (x_idx, y_idx)  # type: ignore[reportAssignmentType] overloaded tuple size in affine
        return x_coords

    def get_cell_y_coords(self, n_rows: int) -> NDArray:
        """Return an array of y coordinates for the center of each cell.

        The coordinates will be in the coordinate system defined by the
        raster's transform.

        Args:
            n_rows: Number of rows in the raster array.

        Returns:
            y_coordinates at cell centers, with shape (n_rows,)
        """
        x_idx = np.zeros(n_rows)  # Use x=0 for a single column
        y_idx = np.arange(n_rows) + 0.5
        _, y_coords = self.transform * (x_idx, y_idx)  # type: ignore[reportAssignmentType] overloaded tuple size in affine
        return y_coords
